import logging

import asyncpg
from pydantic import BaseModel

from settings import SETTINGS

logger = logging.getLogger(__name__)


class Variable(BaseModel):
    key: str
    value: str
    template: bool


class VarService:
    def __init__(
            self,
            connection: str = None
    ):
        self.connection = connection or SETTINGS.prompt_admin_settings.var_connection

    async def collect(self) -> list[Variable]:
        try:
            conn = await asyncpg.connect(self.connection)
        except Exception as e:
            logger.error('Error connection database for get vars', exc_info=e)
            return {}

        row = await conn.fetch('SELECT key, value, template FROM pa_var')

        return [
            Variable(
                key=i.get('key'),
                value=i.get('value'),
                template=i.get('template')
            )
            for i in row
        ]

    async def collect_vars(self) -> dict[str, str]:
        try:
            conn = await asyncpg.connect(self.connection)
        except Exception as e:
            logger.error('Error connection database for get vars', exc_info=e)
            return {}

        row = await conn.fetch('SELECT key, value FROM pa_var WHERE template = FALSE')

        return {
            i.get('key'): i.get('value') for i in row
        }

    async def collect_templates(self) -> dict[str, str]:
        try:
            conn = await asyncpg.connect(self.connection)
        except Exception as e:
            logger.error('Error connection database for get vars', exc_info=e)
            return {}

        row = await conn.fetch('SELECT key, value FROM pa_var WHERE template = TRUE')

        return {
            i.get('key'): i.get('value') for i in row
        }

    async def get_var(self, key: str) -> str:
        try:
            conn = await asyncpg.connect(self.connection)
        except Exception as e:
            logger.error('Error connection database for get vars', exc_info=e)
            raise ValueError()

        row = await conn.fetchrow('SELECT value FROM pa_var WHERE key=$1', key)
        var = row.get('value')
        if var is None:
            raise ValueError()
        return var

    async def create(self, key: str, value: str, template: bool):
        try:
            conn = await asyncpg.connect(self.connection)
        except Exception as e:
            logger.error('Error connection database for get vars', exc_info=e)
            return

        await conn.fetch('INSERT INTO pa_var (key, value, template) VALUES ($1, $2, $3)', key, value, template)

    async def change(self, key: str, value: str):
        try:
            conn = await asyncpg.connect(self.connection)
        except Exception as e:
            logger.error('Error connection database for get vars', exc_info=e)
            return

        await conn.fetch('UPDATE pa_var SET value=$1 WHERE key=$2', value, key)

    async def remove(self, key: str):
        try:
            conn = await asyncpg.connect(self.connection)
        except Exception as e:
            logger.error('Error connection database for get vars', exc_info=e)
            return

        await conn.fetch('DELETE FROM pa_var WHERE key=$1', key)



import inspect
import logging
from typing import Callable, Awaitable, Any, get_args

from jinja2.nativetypes import NativeEnvironment
from polyfactory.factories.pydantic_factory import ModelFactory
from pydantic import BaseModel

from promptadmin.output.parser_output_service import ParserOutputService
from promptadmin.prompt_service.models.base_model_service import BaseModelService, Message, ModelResponse
from promptadmin.types import PromptServiceInfo
from promptadmin.vars.var_service import VarService

logger = logging.getLogger(__name__)


class InspectPromptService:
    def __init__(
            self,
            table: str,
            field: str,
            field_name: str,
            name: str | None,
            func: Callable[[Any, Any], Awaitable[Any]],
            model_service: BaseModelService,
            template_context_default: dict[str, BaseModel],
            parser_output_service: ParserOutputService = None,
            history_context_default: list[Message] = None,
            parsed_model_default: dict = None,
            fail_parse_model_strategy: str = None,
            var_service: VarService = None
    ):
        self.table = table
        self.field = field
        self.field_name = field_name
        self.name = name
        self.func = func
        self.model_service = model_service
        self.template_context_type = {}
        self.template_context_default = template_context_default
        self.parsed_model_type = None
        self.parser_output_service = parser_output_service or ParserOutputService()
        self.history_context_default = history_context_default or []
        self.parsed_model_default = parsed_model_default
        self.fail_parse_model_strategy = fail_parse_model_strategy
        self.var_service = var_service or VarService()

        self.collect_template_context_type()
        self.collect_parsed_model_type()

    def info(self) -> PromptServiceInfo:
        parsed_model_default = self.parsed_model_default
        if parsed_model_default is None and self.parsed_model_type:
            parsed_model_default = ModelFactory.create_factory(self.parsed_model_type).build().model_dump()
        return PromptServiceInfo(
            table=self.table,
            field=self.field,
            field_name=self.field_name,
            name=self.name,
            service_model_info=self.model_service.info(),
            template_context_type=dict(
                (k, v.model_json_schema(mode='serialization')) for k, v in self.template_context_type.items()
            ),
            template_context_default={
                **{
                    k: v.dict() for k, v in self.template_context_default.items()
                },
                **{
                    k: ModelFactory.create_factory(v).build().model_dump() for k, v in
                    self.template_context_type.items()
                    if k not in self.template_context_default
                }
            },
            history_context_default=self.history_context_default,
            parsed_model_type=self.parsed_model_type.model_json_schema(
                mode='serialization') if self.parsed_model_type else None,
            parsed_model_default=parsed_model_default,
            fail_parse_model_strategy=self.fail_parse_model_strategy
        )

    def collect_template_context_type(self):
        for k, v in inspect.getfullargspec(self.func).annotations.items():
            if k == 'return' or k == 'history':
                continue
            if issubclass(v, BaseModel):
                self.template_context_type[k] = v

    def collect_parsed_model_type(self):
        for k, v in inspect.getfullargspec(self.func).annotations.items():
            if k != 'return':
                continue

            return_type = get_args(v.model_fields['parsed_model'].annotation)[0]
            if str(return_type) == '~T':
                continue
            self.parsed_model_type = return_type

    async def process(
            self,
            *args,
            prompt: str | None = None,
            history: list[Message] | None = None,
            mapping_name: str | None = None,
            **kwargs
    ) -> ModelResponse:
        """

        :param args:
        :param prompt:
        :param history:
        :param mapping_name: For logging
        :param kwargs:
        :return:
        """
        contexts = self._collect_contexts(*args, **kwargs)
        prompt = await self._collect_prompt(prompt, contexts)
        history = history or []
        return await self._execute_prompt(prompt, history, mapping_name)

    def _collect_contexts(self, *args, **kwargs):
        contexts = {}
        for v in args:
            for k, v1 in self.template_context_type.items():
                if isinstance(v, v1):
                    contexts[k] = v

        for k, v in kwargs.items():
            for k1, v1 in self.template_context_type.items():
                if isinstance(v, v1):
                    if k == k1:
                        contexts[k] = v

        if len(self.template_context_type) != len(contexts):
            raise ValueError()
        return contexts

    async def _collect_prompt(self, prompt: str | None, contexts) -> str:
        if prompt is None:
            if self.name is None:
                raise ValueError()
            raise NotImplementedError()

        environment = NativeEnvironment(autoescape=True, enable_async=True)
        template_ = environment.from_string(prompt)

        var = await self._collect_vars()

        contexts.update(
            {'var': var}
        )

        async def render_template(template_key: str):
            templ = await self.var_service.get_var(template_key)
            templ_ = environment.from_string(templ)
            return await templ_.render_async(
                **contexts,
                render_template=render_template
            )

        return await template_.render_async(
            **contexts,
            render_template=render_template
        )

    async def _collect_vars(self) -> dict[str, str]:
        return await self.var_service.collect_vars()

    async def _collect_templates(self) -> dict[str, str]:
        return await self.var_service.collect_templates()

    async def _execute_prompt(
            self,
            prompt: str,
            history: list[Message],
            mapping_name: str | None = None,
    ) -> ModelResponse:
        model_response = await self.model_service.execute(prompt, history)

        try:
            if self.parsed_model_type:
                model_response.parsed_model = self.parser_output_service.parse(
                    self.parsed_model_type,
                    model_response.raw_text
                )
                if model_response.parsed_model is None:
                    if self.fail_parse_model_strategy == 'exception':
                        raise ValueError()
                    elif self.fail_parse_model_strategy == 'default':
                        model_response.parsed_model = self.parsed_model_default
        finally:
            logger.info(
                'Model processing',
                extra={
                    'prompt': prompt,
                    'history': history,
                    'mapping': {
                        'table': self.table,
                        'field': self.field,
                        'field_name': self.field_name,
                        'name': mapping_name or self.name,
                    },
                    'model': {
                        'config': self.model_service.info().model_dump(),
                        'response': model_response.model_dump(),
                    },
                    'parsing': {
                        'parsed_model_type': str(self.parsed_model_type),
                        'raw_text': str(model_response.raw_text),
                        'parsed_model': str(model_response.parsed_model)
                    }
                }
            )

        return model_response



import json
import os.path


class TestLoadService:
    @staticmethod
    async def list() -> dict[str, float]:
        if not os.path.exists('test_reports'):
            return {}
        response = {}
        for i in os.listdir('test_reports'):
            try:
                with open(f'test_reports/{i}', 'r') as file:
                    file_json = json.loads(file.read())
                    response[i.removesuffix('.json')] = file_json['created']
            finally:
                pass
        return response

    @staticmethod
    async def load(uuid: str) -> dict | None:
        if not os.path.exists('test_reports'):
            return
        try:
            with open(f'test_reports/{uuid}.json', 'r') as file:
                return json.loads(file.read())
        finally:
            pass



import os

from pydantic import BaseModel


class PromptAdminSettings(BaseModel):
    router_secret: str = os.getenv('PROMPT_ADMIN_ROUTER_SECRET')
    var_connection: str = os.getenv('PROMPT_ADMIN_VAR_CONNECTION')
    test_rerun_timeout: int = 60 * 60 * 6


class Settings(BaseModel):
    prompt_admin_settings: PromptAdminSettings = PromptAdminSettings()



import asyncio

from promptadmin.sync.status_monitor.test_run_service import TestRunService
from settings import SETTINGS


class TestRunJob:
    def __init__(self, test_run_service: TestRunService = None):
        self.test_run_service = test_run_service or TestRunService()

    async def start(self):
        await asyncio.sleep(20)
        while True:
            self.test_run_service.run_test()
            await asyncio.sleep(SETTINGS.prompt_admin_settings.test_rerun_timeout)

    async def stop(self):
        pass



from promptadmin.prompt_service.decorator import instances
from promptadmin.types.project_promt_info import ProjectPromptInfo
from settings import SETTINGS


def collect() -> ProjectPromptInfo:
    return ProjectPromptInfo(
        app=SETTINGS.app,
        prompt_service_info=[i.info() for i in instances]
    )



import logging
from typing import Literal, Generic

from anthropic import AsyncAnthropic, NOT_GIVEN
from anthropic.types.message import Message as AnthropicMessage

from promptadmin.prompt_service.models.abstract_streamer import AbstractStreamer
from promptadmin.prompt_service.models.base_model_service import BaseModelService
from promptadmin.types import ModelResponse, Message, ModelServiceInfo
from promptadmin.types.model_response import T
from settings import SETTINGS

logger = logging.getLogger(__name__)


class AnthropicModelResponse(ModelResponse[T], Generic[T]):
    origin_message: AnthropicMessage | None = None
    messages: list[dict[str, str]]


class AnthropicModelService(BaseModelService):
    def __init__(
            self,
            anthropic: AsyncAnthropic = None,
            prompt_position: Literal['system', 'user'] = None,
            max_tokens: int = None,
            model: str = None,
            temperature: float = None
    ):
        self.anthropic = anthropic or AsyncAnthropic(api_key=SETTINGS.anthropic_key)
        self.prompt_position = prompt_position or 'system'
        self.max_tokens = max_tokens or 2048
        self.model = model or 'claude-3-5-sonnet-20240620'
        self.temperature = temperature or NOT_GIVEN

    def info(self) -> ModelServiceInfo:
        return ModelServiceInfo(
            service='anthropic',
            model=self.model,
            config={
                'max_tokens': self.max_tokens,
                'prompt_position': self.prompt_position,
                'temperature': None if self.temperature is NOT_GIVEN else self.temperature
            }
        )

    @staticmethod
    def from_info(model_service_info: ModelServiceInfo):
        return AnthropicModelService(
            prompt_position=model_service_info.config.get('prompt_position'),
            max_tokens=model_service_info.config.get('max_position'),
            model=model_service_info.model,
            temperature=model_service_info.config.get('temperature'),
        )

    async def execute(
            self,
            prompt: str,
            history: list[Message],
            streamer: AbstractStreamer = None
    ) -> AnthropicModelResponse:
        messages = []
        system = NOT_GIVEN
        if self.prompt_position == 'user':
            history.append(Message(content=prompt, role='user'))
        elif self.prompt_position == 'system':
            system = prompt
            if len(history) == 0:
                history.append(Message(content="Hello", role='user'))

        for message in history:
            if message.role == 'user':
                if len(messages) > 0 and messages[-1]['role'] == 'user':
                    messages[-1]['content'] += '\n\n' + message.content
                else:
                    messages.append({'role': 'user', 'content': message.content})
            if message.role == 'assistant':
                if len(messages) <= 0:
                    messages.append({'role': 'user', 'content': 'Hello'})
                if messages[-1]['role'] == 'assistant':
                    messages[-1]['content'] += '\n\n' + message.content
                else:
                    messages.append({'role': 'assistant', 'content': message.content})

        if streamer:
            async with self.anthropic.messages.stream(
                    max_tokens=self.max_tokens,
                    messages=messages,
                    system=system,
                    model=self.model,
                    temperature=self.temperature
            ) as stream:
                async for text in stream.text_stream:
                    await streamer.stream(text)
                anthropic_message = await stream.get_final_message()
        else:
            anthropic_message = await self.anthropic.messages.create(
                max_tokens=self.max_tokens,
                messages=messages,
                system=system,
                model=self.model,
                temperature=self.temperature
            )

        return AnthropicModelResponse(
            raw_text=anthropic_message.content[0].text,
            origin_message=anthropic_message,
            messages=messages
        )



from pydantic import BaseModel

from promptadmin.types import ModelServiceInfo, Message


class PromptServiceInfo(BaseModel):
    table: str
    field: str
    field_name: str
    name: str | None
    service_model_info: ModelServiceInfo

    template_context_type: dict[str, dict]
    template_context_default: dict[str, dict]

    history_context_default: list[Message] = []

    parsed_model_type: dict | None = None
    parsed_model_default: dict | None = None
    fail_parse_model_strategy: str | None = None



import os
import subprocess
import uuid

_process = False


class TestRunService:
    def __init__(self):
        if not os.path.exists('test_reports'):
            os.mkdir('test_reports')

    def run_test(self):
        global _process
        if _process:
            return
        _process = True
        try:
            subprocess.Popen(['pytest', '--json-report', f'--json-report-file=test_reports/{uuid.uuid4()}.json'])
        finally:
            _process = False



from __future__ import annotations
from abc import abstractmethod

from promptadmin.prompt_service.models.abstract_streamer import AbstractStreamer
from promptadmin.types import Message, ModelResponse, ModelServiceInfo


class BaseModelService:
    @abstractmethod
    async def execute(
            self,
            prompt: str,
            history: list[Message],
            streamer: AbstractStreamer = None
    ) -> ModelResponse:
        """"""

    @abstractmethod
    def info(self) -> ModelServiceInfo:
        """"""



from typing import Awaitable, Callable, Any, Literal

from pydantic import BaseModel

from promptadmin.prompt_service.inspect_prompt_service import InspectPromptService
from promptadmin.prompt_service.models.anthropic import AnthropicModelService
from promptadmin.prompt_service.models.base_model_service import BaseModelService, ModelResponse
from promptadmin.types import Message

instances: list[InspectPromptService] = []


def bind_prompt(
        table: str,
        field: str,
        field_name: str,
        name: str | None = None,
        model_service: BaseModelService | None = None,
        template_context_default: dict[str, BaseModel] | None = None,
        history_context_default: list[Message] = None,
        parsed_model_default: BaseModel | dict = None,
        fail_parse_model_strategy: Literal['exception', 'default'] | None = None
):
    model_service = model_service or AnthropicModelService()
    template_context_default = template_context_default or {}
    history_context_default = history_context_default or []

    def bind_prompt_wrapper(func: Callable[[Any, Any], Awaitable[Any]]):
        inspect_prompt_service = InspectPromptService(
            table,
            field,
            field_name,
            name,
            func,
            model_service=model_service,
            template_context_default=template_context_default,
            history_context_default=history_context_default,
            parsed_model_default=parsed_model_default,
            fail_parse_model_strategy=fail_parse_model_strategy
        )
        instances.append(inspect_prompt_service)

        async def wrapper(
                self,
                *args,
                prompt: str | None = None,
                history: list[Message] | None = None,
                **kwargs
        ) -> ModelResponse:
            return await inspect_prompt_service.process(*args, prompt=prompt, history=history, **kwargs)

        return wrapper

    return bind_prompt_wrapper



import uuid
from typing import Generic, TypeVar, Union

from pydantic import BaseModel
from sqlmodel import Field


ParsedModelType = BaseModel
T = TypeVar('T', bound=ParsedModelType)


class ModelResponse(BaseModel, Generic[T]):
    uuid: str = Field(default_factory=lambda: str(uuid.uuid4()))
    raw_text: str
    parsed_model: Union[T, None] = None



from fastapi import APIRouter, Request, BackgroundTasks
from pydantic import BaseModel

from promptadmin.sync.collector import collect
from promptadmin.sync.status_monitor.test_load_service import TestLoadService
from promptadmin.sync.status_monitor.test_run_service import TestRunService
from promptadmin.types import ProjectPromptInfo
from settings import SETTINGS

router = APIRouter(prefix='/prompt-admin')

test_run_service = TestRunService()
test_load_service = TestLoadService()


class TestCaseResponse(BaseModel):
    test_case: dict | None


@router.get('/collect', response_model=ProjectPromptInfo)
async def get_collect(request: Request):
    if request.headers.get('Prompt-Admin-Secret', None) != SETTINGS.prompt_admin_settings.router_secret:
        raise ValueError()
    return collect()


@router.get('/status/run_test')
async def status_run_test(request: Request, background_tasks: BackgroundTasks):
    if request.headers.get('Prompt-Admin-Secret', None) != SETTINGS.prompt_admin_settings.router_secret:
        raise ValueError()
    background_tasks.add_task(test_run_service.run_test)


@router.get('/status/test_list', response_model=dict[str, float])
async def status_test_list(request: Request):
    if request.headers.get('Prompt-Admin-Secret', None) != SETTINGS.prompt_admin_settings.router_secret:
        raise ValueError()
    return await test_load_service.list()


@router.get('/status/test/{uuid}', response_model=TestCaseResponse)
async def status_test_load(request: Request, uuid: str):
    if request.headers.get('Prompt-Admin-Secret', None) != SETTINGS.prompt_admin_settings.router_secret:
        raise ValueError()
    return TestCaseResponse(
        test_case=await test_load_service.load(uuid)
    )



from .anthropic import AnthropicModelService
from .anthropic_bedrok import AnthropicBedrockModelService
from .base_model_service import BaseModelService
from ...types import ModelServiceInfo


def build_model(model_service_info: ModelServiceInfo) -> BaseModelService:
    if model_service_info.service == 'anthropic':
        return AnthropicModelService.from_info(model_service_info)
    elif model_service_info.service == 'anthropic_bedrock':
        return AnthropicBedrockModelService.from_info(model_service_info)

    raise ValueError()


__all__ = [
    AnthropicModelService,
    build_model
]



from .model_response import ModelResponse
from .model_service_info import ModelServiceInfo
from .message import Message
from .prompt_service_info import PromptServiceInfo
from .settings import Settings, PromptAdminSettings
from .project_promt_info import ProjectPromptInfo

__all__ = [
    ModelResponse,
    ModelServiceInfo,
    Message,
    PromptServiceInfo,
    Settings,
    PromptAdminSettings,
    ProjectPromptInfo
]



import json
import os
import re
from json import JSONDecodeError
import os.path
from pathlib import Path

from datamodel_code_generator import generate, InputFileType, PythonVersion
import importlib
from tempfile import NamedTemporaryFile
import xmltodict
from pydantic import BaseModel, ValidationError


class ParserOutputService:
    def parse(self, model_type: type[BaseModel], result: str) -> BaseModel | None:
        try:
            return model_type.parse_obj(self.try_extract_json(result))
        except StopIteration:
            pass
        except JSONDecodeError:
            pass
        except ValidationError:
            pass
        try:
            return model_type.parse_obj(self.try_parse_xml(result))
        except ValidationError:
            pass

    def parse_for_json_schema(self, json_schema: dict, result: str) -> dict | None:
        model = self.load_model_from_schema(json_schema)
        if model:
            return self.parse(model, result)

    @staticmethod
    def load_model_from_schema(schema: dict) -> type(BaseModel) | None:
        title = schema['title']

        if not os.path.exists('_models'):
            os.mkdir('_models')
        file_model = NamedTemporaryFile(dir='_models', suffix='.py')

        generate(
            str(schema),
            input_file_type=InputFileType.JsonSchema,
            output=Path(file_model.name),
            target_python_version=PythonVersion.PY_311,
        )

        module = importlib.import_module(f'_models.{file_model.name.split("/")[-1].removesuffix(".py")}')
        model = getattr(module, title)
        return model

    @staticmethod
    def try_extract_json(s: str):
        if '{' not in s and '[' not in s:
            s = '{' + s
        s = s[next(idx for idx, c in enumerate(s) if c in '{['):]
        try:
            return json.loads(s)
        except json.JSONDecodeError as e:
            return json.loads(s[:e.pos])

    @staticmethod
    def try_parse_xml(s: str):
        s = s.replace('\n', '@||')
        tag = re.search(r'<([a-zA-Z_]*)>', s)
        if tag:
            first_tag = tag.group(1)
            res = re.search(rf'<{first_tag}>.*</{first_tag}>', s)
            if res:
                find = res.group(0)
                find = find.replace('@||', '\n')
                return xmltodict.parse(find, attr_prefix='attr_')
        res = re.search(r'<[^>]*/>', s)
        if res is None:
            return None
        find = res.group(0)
        find = find.replace('@||', '\n')
        return xmltodict.parse(find, attr_prefix='attr_')
